/*---------------------------------------------------------------------------*/

/*!  \file  simulator.c
     \brief  a simple brownian dynamics simulator.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2007,2008,2020 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif
#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#include "condition.h"
#include "iofiles.h"
#include "geometry.h"
#include "blend.h"
#include "particles.h"
#include "textwrite.h"
#include "force.h"
#include "integrator.h"
#include "celllist.h"
#include "dxwrite.h"

/*---------------------------------------------------------------------------*/
static int filename_is_valid(char *filename);
static void write_data_to_file(simulation_condition *condition,
                               simulation_iofiles *iofiles,int n,
                               soft_particles *particles,
                               linked_cell_list *celllist,
                               polymer_blend *blend,
                               box_geometry *geometry);

/*---------------------------------------------------------------------------*/
int simulate(simulation_condition *condition,
             simulation_iofiles *iofiles,
             box_geometry *geometry,
             polymer_blend *blend,
             soft_particles *particles)
{
    int i,n;
    linked_cell_list *celllist;

    if((geometry->dimension != 2) && (geometry->dimension != 3))
    {
        fprintf(stderr,
                "%d dimensional simulations are not supported (only 2 and 3 are supported).",
                geometry->dimension);
        return EXIT_FAILURE;
    }
    fprintf(stderr,"dimensions: %d\n",geometry->dimension);
    fprintf(stderr,"total number of particles: %d\n",
            total_number_of_soft_particles(particles));
    fprintf(stderr,"spatial average of total particle density (%dD): %g\n",
            geometry->dimension,
            total_number_of_soft_particles(particles)
            / (geometry->l[0] * geometry->l[1] * geometry->l[2]));
    if(condition->apply_shear)
    {
        if(condition->omega > 0.0)
            fprintf(stderr,"shear flow: oscillatory (kappa: %g, omega: %g)\n",
                    condition->kappa,condition->omega);
        else
            fprintf(stderr,"shear flow: steady (kappa: %g)\n",
                    condition->kappa);
    }
    else
    {
        fprintf(stderr,"shear flow: no \n");
    }
    fprintf(stderr,"time step: %g\n",condition->dt);


    if(condition->interval <= 0)
        condition->interval = condition->iteration_max;

    if(condition->apply_shear)
    {
        geometry->kappa = condition->kappa;
        geometry->omega = condition->omega;
        if(geometry->omega < 0.0)
            geometry->omega = 0.0;
    }
    else
    {
        geometry->kappa = 0.0;
        geometry->omega = 0.0;
    }

    geometry->delta_l = 0.0;
    geometry->theta = 0.0;

    calculate_lambda0(blend,geometry);

    celllist = allocate_linked_cell_list(particles,geometry);


        /*  main simulation loop.  */
    n = 0;

    for(i = 0;i <= condition->iteration_max;i++)
    {
        integrate_langevin_equation(particles,celllist,blend,geometry,
                                    condition->dt);

        /* if(((i % condition->interval) == 0) && (i > 0)) */
        if((i % condition->interval) == 0)
        {
            write_data_to_file(condition,iofiles,n,
                               particles,celllist,blend,geometry);
            n++;
        }
    }

    /* if((i % condition->interval) == 0) */
    /*     write_data_to_file(condition,iofiles,n, */
    /*                        particles,celllist,blend,geometry); */


    free_linked_cell_list(celllist);


    fprintf(stderr,"number of total iterations: %d\n",i);

    return EXIT_SUCCESS;
}

/*---------------------------------------------------------------------------*/
void write_data_to_file(simulation_condition *condition,
                        simulation_iofiles *iofiles,int n,
                        soft_particles *particles,linked_cell_list *celllist,
                        polymer_blend *blend,box_geometry *geometry)
{
    static char filename[FILENAME_BUFFERSIZE];

    if(condition->save_position_sequential)
    {
        sprintf(filename,iofiles->position_template,n);
        write_positions_text(filename,particles);
    }
    else if(filename_is_valid(iofiles->position_output))
    {
        sprintf(filename,"%s~",iofiles->position_output);
        write_positions_text(filename,particles);
        rename(filename,iofiles->position_output);
    }

    /* if(condition->save_stress_sequential) */
    /* { */
    /*     sprintf(filename,iofiles->stress_template,n); */
    /*     calculate_stress_tensor(particles,celllist, */
    /*                             blend,geometry); */
    /*     write_stresses_text(filename,particles,geometry); */
    /* } */
    /* else */
    if(filename_is_valid(iofiles->stress_output))
    {
        if(condition->save_stress_sequential)
            strcpy(filename,iofiles->stress_output);
        else
            sprintf(filename,"%s~",iofiles->stress_output);

        calculate_stress_tensor(particles,celllist,
                                blend,geometry);
        write_stresses_text(filename,particles,geometry,
                            condition->save_stress_sequential);

        if(!(condition->save_stress_sequential))
            rename(filename,iofiles->stress_output);
    }

    if(condition->save_dx_sequential)
    {
        sprintf(filename,iofiles->dx_template,n);
        write_positions_dx(filename,particles,geometry);
    }
    else if(filename_is_valid(iofiles->dx_output))
    {
        sprintf(filename,"%s~",iofiles->dx_output);
        write_positions_dx(filename,particles,geometry);
        rename(filename,iofiles->dx_output);
    }

    /* if(condition->save_geometry_sequential) */
    /* { */
    /*     sprintf(filename,iofiles->geometry_template,n); */
    /*     write_box_geometry_text(filename,geometry); */
    /* } */
    /* else */
    if(filename_is_valid(iofiles->geometry_output))
    {
        if(condition->save_geometry_sequential)
            strcpy(filename,iofiles->geometry_output);
        else
            sprintf(filename,"%s~",iofiles->geometry_output);

        write_box_geometry_text(filename,geometry,
                                condition->save_geometry_sequential);

        if(!(condition->save_geometry_sequential))
            rename(filename,iofiles->geometry_output);
    }
}

/*---------------------------------------------------------------------------*/
int filename_is_valid(char *filename)
{
    if((filename == NULL) || (strlen(filename) == 0))
    {
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

/*---------------------------------------------------------------------------*/
